/**
   <title>AGSTex ... an output class producing TeX for autogsdoc</title>
   Copyright (C) 2001 Willem Rein Oudshoorn.

   Written by: Willem Rein Oudshoorn <woudshoo@xs4all.nl>
   Created: December 2001

   This file will be part of the GNUstep Project

   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   You should have received a copy of the GNU General Public
   License along with this program; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <Foundation/Foundation.h>
#include "AGSTex.h"

@class FileManager;

extern FileManager *theFileManager;

@interface AGSTex (Private)

- (void) _output_arg_without_comma: (GSXMLNode*) node;

@end


@implementation AGSTex

/**
   Put the default entires in the prefix/postfix stack.  This
   is usefull as it avoids writing methods for simple cases.
   For example, this has entries mapping the gsdoc node
   <code>declared</code> to the pair <code>"\n\edeclaredInFile\001"</code>
   and <code>"\002"</code>
*/
- (void) initPrefixPostfixStack
{
    [self addLayerToStack];

    [self setPrefix: @"\eselectorpart\001" andPostfix: @"\002" forEntity: @"sel"];
    [self setPrefix: @"\n\ebeginDescription " andPostfix: @"\eendDescription " forEntity: @"desc"];
    [self setPrefix: @"\n\edeclaredInFile\001" andPostfix: @"\002" forEntity: @"declared"];
    [self setPrefix: @"\eimplementsProtocol\001" andPostfix: @"\002\n" forEntity: @"conform"];
    [self setPrefix: @"\n\ebeginList\n" andPostfix: @"\n\eendList\n" forEntity: @"list"];
    [self setPrefix: @"\n\eterm\001" andPostfix: @"\002" forEntity: @"term"];
    [self setPrefix: @"\n\eitem " forEntity: @"item"];
    [self setPrefix: @"\ecode\001" andPostfix: @"\002" forEntity: @"code"];
    [self setPrefix: @"\ecode\001" andPostfix: @"\002" forEntity: @"file"];
    [self setPrefix: @"\ebold\001" andPostfix: @"\002" forEntity: @"strong"];
    [self setPrefix: @"\eitalics\001" andPostfix: @"\002" forEntity: @"em"];
    [self setPrefix: @"\eitalics\001" andPostfix: @"\002" forEntity: @"var"]; 
    [self setPrefix: @"\n\n" forEntity: @"br"];
    [self setPrefix: @"\n\n" forEntity: @"p"];
    [self setPrefix: @"\ecopyrightSection\001" andPostfix: @"\002\n" forEntity: @"copy"];
    [self setPrefix: @"\005 Start body\n" andPostfix: @"\n\005 end body\n" forEntity: @"body"];
    [self setPrefix: @"\005 start back\n" andPostfix: @"\n\005 end back\n" forEntity: @"back"];
    [self setPrefix: @"\005 Here should come a table of contents\n"
          andPostfix: @"\005 end table of contents\n" forEntity: @"contents"];
    [self setPrefix: @"\005 front matter, like table of contents\n" andPostfix: @"\005 end of front matter\n" forEntity: @"front"];
    [self setPrefix: @"\etitle\001" andPostfix: @"\002\n" forEntity: @"title"];
    [self setPrefix: @"\eversion\001" andPostfix: @"\002" forEntity: @"version"];

    [self setPrefix: @"\estandards\001" andPostfix: @"\002" forEntity: @"standards"];
    [self setPrefix: @"\estandardGNU" forEntity: @"GNUstep"];
    [self setPrefix: @"\estandardOS" forEntity: @"OpenStep"];
    [self setPrefix: @"\estandardNOS" forEntity: @"NotOpenStep"];
    [self setPrefix: @"\estandardNOSX" forEntity: @"NotMacOS-X"];
    [self setPrefix: @"\estandardOSX" forEntity: @"MacOS-X"];
}

- init
{
    self = [super init];
    
    RELEASE (entityMapping);
    
    entityMapping = [[NSDictionary alloc] initWithObjectsAndKeys:
                                              @"<", @"lt",
                                          @">", @"gt",
                                          @"\ecopyright\001\002", @"copy",
                                          @"(tm)", @"tm",  // TODO fix this
                                          @"\edots\001\002", @"dots",
                                          @"\epenalty10000\e ", @"nbsp", 
                                          @"&", @"amp", 
                                          @"''", @"quot", // TODO check this
                                          @"(reg)", @"reg", // TODO fix this
                                          nil, nil];
    _currentClass = @"";

    [self initPrefixPostfixStack];
    
    return self;
}

- (void) dealloc
{
    RELEASE (_currentClass);

    [super dealloc];
}

- (void) _output_gsdoc: (GSXMLNode*) node
{
    NSString *texMacroFile;

    texMacroFile = [[NSBundle mainBundle] pathForResource: @"gsdocmacros"
                                          ofType: @"tex"];
    if (texMacroFile == nil)
        {
        texMacroFile = @"gsdocmacros.tex";
        }
    
    [buffer appendString: @"%% begin of gsdoc node\n"];
    [buffer appendString: @"\\catcode`\\^^[=0 %% escape character is \\e\n"];
    [buffer appendString: @"\\catcode`\\^^A=1 %% Ctrl-A is open group\n"];
    [buffer appendString: @"\\catcode`\\^^B=2 %% Ctrl-B is close group\n"];
    [buffer appendString: @"\\catcode`\\^^D=4 %% Ctrl-D is the alignment character\n"];
    [buffer appendString: @"\\catcode`\\^^E=14 %% Ctrl-F is the comment character\n"];
    [buffer appendFormat: @"\einput %@\n", texMacroFile];
    [buffer appendString: @"\ecatcode`\e\\=12 %% \\ is a normal character\n"];
    [buffer appendString: @"\ecatcode`\e^=12 %% ^ is a normal character\n"];
    [buffer appendString: @"\ecatcode`\e~=12 %% ~ is a normal character\n"];
    [buffer appendString: @"\ecatcode`\e{=12 %% { is a normal character\n"];
    [buffer appendString: @"\ecatcode`\e}=12 %% } is a normal character\n"];
    [buffer appendString: @"\ecatcode`\e&=12 %% & is a normal character\n"];
    [buffer appendString: @"\ecatcode`\e#=12 %% # is a normal character\n"];
    [buffer appendString: @"\ecatcode`\e%=12 \005 %% is the normal character\n"];
    [buffer appendString: @"\ecatcode`\e$=12 \005 %% $ is a normal character\n"];
    [buffer appendString: @"\echardef\e\\=`\e\\ \005 \\ is the character `\\'\n"];
    [buffer appendString: @"\echardef\e&=`\e& \005 & is the character `&'\n"];
    [buffer appendString: @"\echardef\e#=`\e# \005 \n"];
    [self outputNodeList: [node firstChild]]; 
    [buffer appendString: @"\eend\n"];
    [buffer appendString: @"\n\005 gsdoc node close\n"];
}


- (void) _output_head: (GSXMLNode*) node
{
    NSMutableDictionary *headNodes = [self collectChildNodesByName: node];

    [buffer appendString: @"\eexpandafter\eifx\ecsname included\eendcsname\erelax\n"];
    [buffer appendString: @"\ebeginHead\n"];

    [self outputNodesOf: headNodes forKeyAndRemoveKey: @"title"];
    [self outputNodesOf: headNodes forKeyAndRemoveKey: @"date"];
    if ([headNodes objectForKey: @"author"])
        {
        [buffer appendString: @"\eauthorSection\005\n"];
        [self outputNodesOf: headNodes forKeyAndRemoveKey: @"author"];
        [buffer appendString: @"\n\eendAuthorSection\n"];
        }
    [self outputNodesOf: headNodes forKeyAndRemoveKey: @"copy"];
    [self outputNodesOf: headNodes forKeyAndRemoveKey: @"version"];
    [self outputNodesOf: headNodes forKeyAndRemoveKey: @"abstract"];    
    [self outputNodesOf: headNodes];
    
    [buffer appendString: @"\n\eendHead\n"];
    [buffer appendString: @"\efi\n"];
}


- (void) _output_text: (GSXMLNode*) node
{
    [buffer appendString: [node content]];
}

- (void) _output_author: (GSXMLNode*) node
{
    [buffer appendString: @"\ebeginAuthor\001"];
    [buffer appendString: [node objectForKey: @"name"]];
    [buffer appendString: @"\002"];
    [self outputNodeList: [node firstChild]]; 
    [buffer appendString: @"\eendAuthor"];
}

- (void) _output_email: (GSXMLNode*) node
{
    [buffer appendString: @"\eemail\001"];
    [buffer appendString: [node objectForKey: @"address"]];
    [buffer appendString: @"\002\001"];
    if ([node firstChild])
        {
        [self outputNodeList: [node firstChild]];
        }
    else
        {
        [buffer appendString: [node objectForKey: @"address"]];
        }
    [buffer appendString: @"\002\n"];    
}

- (void) _output_chapter: (GSXMLNode*) node
{
    [buffer appendString: @"\005%% chapter\n"];

    [self addLayerToStack];
    [self mapName: @"heading" toName: @"chapter_heading"];
    [self outputNodeList: [node firstChild]]; 
    [buffer appendString: @"\eendChapter\n\005%% end chapter\n"];
    [self removeLayerFromStack];
}

- (void) _output_chapter_heading: (GSXMLNode*) node
{
    NSMutableString *oldBuffer = buffer;
    buffer = (NSMutableString*) [NSMutableString string];

    [self outputNodeList: [node firstChild]];

    if ([[buffer stringByTrimmingSpaces] hasPrefix: @"Software documentation"])
        {
        buffer = oldBuffer;
        }
    else
        {
        [oldBuffer appendFormat: @"\echapter\001%@\002",  [buffer stringByTrimmingSpaces]];
        buffer = oldBuffer;
        }
}

- (void) _output_section: (GSXMLNode*) node
{
    [buffer appendString: @"\005%% section\n"];
    [self addLayerToStack];
    [self mapName: @"heading" toName: @"heading"];  // This because chapter maps name
    [self setPrefix: @"\esection\001" andPostfix: @"\002\n" forEntity: @"heading"];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"\eendSection\n\005%% end section\n"];
    [self removeLayerFromStack];
}

- (void) outputClassMethods: (NSArray*) methods
{
    if ([methods count])
        {
        [buffer appendString: @"\n\ebeginClassMethodSection\n"];
        [self outputNodeArray: methods];
        [buffer appendString: @"\n\eendClassMethodSection\n"];
        }
}

- (void) outputInstanceMethods: (NSArray*) methods
{
    if ([methods count])
        {
        [buffer appendString: @"\n\ebeginInstanceMethodSection\n"];
        [self outputNodeArray: methods];
        [buffer appendString: @"\n\eendInstanceMethodSection\n"];
        }
}


- (void) _output_subsect: (GSXMLNode*) node
{
    [buffer appendString: @"\005%% subsection\n"];
    [self addLayerToStack];
    [self setPrefix: @"\esubsection\001" andPostfix: @"\002\n" forEntity: @"heading"];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"\eendSubSection\n\005%% end subsection\n"];
    [self removeLayerFromStack];
}

- (void) _output_deflist: (GSXMLNode*) node
{
    [self addLayerToStack];
    [self setPrefix: @"" andPostfix: @"" forEntity: @"desc"];
    [buffer appendString: @"\n\ebeginDescriptionList\n"];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"\n\eendDescriptionList\n"];
    [self removeLayerFromStack];
}

- (void) _output_enum: (GSXMLNode*) node
{
    [self addLayerToStack];
    [self setPrefix: @"\enumberedItem " forEntity: @"item"];
    [buffer appendString: @"\n\ebeginNumberedList\n"];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"\n\eendNumberedList\n"];
    [self removeLayerFromStack];
}


- (void) _output_example: (GSXMLNode*) node
{
    NSMutableString *oldBuffer = buffer;
    NSArray *tmp;
    
    [buffer appendString: @"\ebeginVerbatim "];
    buffer = (NSMutableString*) [NSMutableString string];
    [self outputNodeList: [node firstChild]];
    tmp = [buffer componentsSeparatedByString: @" "];
    buffer = oldBuffer;
    [buffer appendString: [tmp componentsJoinedByString: @"\e "]];
    [buffer appendString: @"\eendVerbatim"];
}

- (void) _output_variable: (GSXMLNode*) node
{
    [self addLayerToStack];
    [self setPrefix: @"\ebeginVariableDescription "
          andPostfix: @"\n\eendVariableDescription\n"
          forEntity: @"desc"];

    [buffer appendString: @"\evariableDefinition"]; // 3 argument 2 is bold, 1, 2, 3; are displayed.
    [buffer appendFormat: @"\001%@\002", [node objectForKey: @"type"]];
    [buffer appendFormat: @"\001%@\002", [node objectForKey: @"name"]];
    [buffer appendFormat: @"\001\002"];
    [self outputNodeList: [node firstChild]];

    [self removeLayerFromStack];
}

- (NSArray*) _mixedSelArgFromNodeDictionary: (NSDictionary*) nodes
{
    NSMutableArray *mixedSelArg = [NSMutableArray array];
    
    NSEnumerator *numS;
    NSEnumerator *numA;
    id selN, argN;

    numS = [[nodes objectForKey: @"sel"] objectEnumerator];
    numA = [[nodes objectForKey: @"arg"] objectEnumerator];

    selN = [numS nextObject];
    argN = [numA nextObject];
    
    while (selN || argN)
        {
        if (selN)
            {
            [mixedSelArg addObject: selN];
            selN = [numS nextObject];
            }
        
        if (argN)
            {
            [mixedSelArg addObject: argN];
            argN = [numA nextObject];
            }
        }

    return mixedSelArg;
}

- (void) _output_function: (GSXMLNode*) node
{
    NSDictionary *childNodes = [self collectChildNodesByName: node];

    [self addLayerToStack];
    [self setPrefix: @"\ebeginFunctionDescription\n"
          andPostfix:@"\n\eendFunctionDescription\n"
          forEntity: @"desc"];

    [self mapName: @"arg" toName: @"function_arg"];
    
    [buffer appendFormat: @"\n\efunctionDefinition\001%@\002\001%@\002\001",
            [node objectForKey: @"type"],
            [node objectForKey: @"name"]];
    [buffer appendFormat: @"\edest\001function**%@\002", [node objectForKey: @"name"]];
    
    if ([[childNodes objectForKey: @"arg"] count] > 0)
        {
        [self _output_arg_without_comma: [[childNodes objectForKey: @"arg"] objectAtIndex: 0]];
        [[childNodes objectForKey: @"arg"] removeObjectAtIndex: 0];
        }
    
    [self outputNodeArray: [childNodes objectForKey: @"arg"]];

    [buffer appendString: @"\002\n"];
    
    [self outputNodeArray: [childNodes objectForKey: @"desc"]];

    [self removeLayerFromStack];
}

- (void) _output_method: (GSXMLNode*) node
{

    NSMutableDictionary *childNodes = [self collectChildNodesByName: node];

    [self addLayerToStack];
    [self setPrefix: @"\ebeginMethodDescription "
          andPostfix: @"\n\eendMethodDescription\n"
          forEntity: @"desc"];

    if ([[node objectForKey: @"factory"]  boolValue])
        {
        [buffer appendString: @"\n\eclassMethodSubSection"];
        }
    else
        {
        [buffer appendString: @"\n\einstanceMethodSubSection"];
        }
    [buffer appendString: @"\001"];
    if ([[node objectForKey: @"init"] boolValue])
        {
        [buffer appendString: @"\edesignatedInitializer"];
        }
    if ([node objectForKey: @"override"])
        {
        [buffer appendFormat: @"\eoverride%@", [node objectForKey: @"override"]];
        }
    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"standards"];
    [buffer appendString: @"\002"];
        
    [self setPrefix: @"" andPostfix: @"" forEntity: @"sel"];
    [buffer appendString: @"\001"];
    [self outputNodeArray: [childNodes objectForKey: @"sel"]];
    [buffer appendString: @"\002"];
    [self setPrefix: @"\eselectorpart\001" andPostfix: @"\002" forEntity: @"sel"];
    [buffer appendFormat: @"\001\etype\001%@\002\002", [node objectForKey: @"type"]];
    [buffer appendString: @"\001"];
    [self outputNodeArray: [self _mixedSelArgFromNodeDictionary: childNodes]];
    [buffer appendString: @"\002"];
    // make reference anchor
    [buffer appendFormat: @"\edest\001method**%@**", _currentClass];
    [buffer appendString: [[node objectForKey: @"factory"] boolValue] ? @"+": @"-"];
    [self setPrefix: @"" andPostfix: @"" forEntity: @"sel"];
    [self outputNodeArray: [childNodes objectForKey: @"sel"]];
    [buffer appendString: @"\002\n\n"];
    [self outputNodeArray: [childNodes objectForKey: @"desc"]];
    [buffer appendString: @"\n\n\eendMethod\n"];
    [self removeLayerFromStack];
}

- (void) _output_arg: (GSXMLNode*) node
{
    [buffer appendFormat: @"\004\earg\001%@\002\001", [node objectForKey: @"type"]];
    [self outputNodeList: [node firstChild]]; // [node children]];
    [buffer appendString: @"\002\ecr\n"];
}

- (void) _output_function_arg: (GSXMLNode*) node
{
    [buffer appendString: @", "];
    [self _output_arg_without_comma: node];
}

- (void) _output_arg_without_comma: (GSXMLNode*) node
{
    [buffer appendFormat: @"\ecode\001%@\002 \eitalics\001",
            [node objectForKey: @"type"]];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"\002"];
}

- (void) _output_class: (GSXMLNode*) node
{
    NSMutableDictionary *classNodes = [self collectChildNodesByName: node];
    
    ASSIGN (_currentClass, [node objectForKey: @"name"]);
    [self addLayerToStack];
    [self setPrefix: @"\ebeginInterfaceDescription "
          andPostfix: @"\n\eendInterfaceDescription\n"
          forEntity: @"desc"];
    
    [buffer appendString: @"\n\ebeginClassSection\001"];
    [self outputNodesOf: classNodes forKeyAndRemoveKey: @"standards"];
    [buffer appendString: @"\002\001"];
    [buffer appendString: [node objectForKey: @"name"]];
    [buffer appendString: @"\002\n"];

    [buffer appendString: @"\n\ebeginInheritanceSection\n"];
    if ([node objectForKey: @"super"])
        {
        NSDictionary *mappingToSuper = [[globalRefs refs] objectForKey: @"super"];
        NSString *current = [mappingToSuper objectForKey: [node objectForKey: @"name"]];
        
        [buffer appendFormat: @"\einheritsFromA\001%@\002\n", current];

        while ((current = [mappingToSuper objectForKey: current]))
            {
            [buffer appendFormat: @"\einheritsFromB\001%@\002\n", current];
            }
        }
    else
        {
        [buffer appendFormat: @"\eisRootClass\n"];
        }
    
    [buffer appendString: @"\n\eendInheritanceSection\n"];
    
    if ([classNodes objectForKey: @"conform"])
        {
        [buffer appendString: @"\ebeginConformSection\n"];
        [self outputNodesOf: classNodes forKeyAndRemoveKey: @"conform"];
        [buffer appendString: @"\eendConformSection\n"];
        }
    [self outputNodesOf: classNodes forKeyAndRemoveKey: @"declared"];
    [self outputNodesOf: classNodes forKeyAndRemoveKey: @"desc"];

    [self outputClassesAndInstanceMethodsFromArray: [classNodes objectForKey: @"method"]];
    [classNodes removeObjectForKey: @"method"];
    
    [self outputNodesOf: classNodes];
    [buffer appendString: @"\n\eendClassSection\n"];
    [self removeLayerFromStack];

    ASSIGN (_currentClass, @"");
}

- (void) _output_protocol: (GSXMLNode*) node
{
    NSMutableDictionary *childNodes = [self collectChildNodesByName: node];
    NSArray *methods;
    
    [self addLayerToStack];
    [self setPrefix: @"\ebeginProtocolDescription "
          andPostfix: @"\n\eendProtocolDescription\n"
          forEntity: @"desc"];
    
    [buffer appendString: @"\n\ebeginProtocolSection\001"];
    [buffer appendString: [node objectForKey: @"name"]];
    [buffer appendString: @"\002\n"];
    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"conform"];
    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"declared"];
    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"desc"];
    
    methods = AUTORELEASE (RETAIN ([childNodes objectForKey: @"method"]));
    [childNodes removeObjectForKey: @"method"];
    [self outputNodesOf: childNodes];
    [self outputClassesAndInstanceMethodsFromArray: methods];
    [buffer appendString: @"\n\eendProtocolSection\n"];
    [self removeLayerFromStack];
}

- (void) _output_category: (GSXMLNode*) node
{
    NSMutableDictionary *childNodes = [self collectChildNodesByName: node];
    NSArray *methods;
    
    [self addLayerToStack];
    [self setPrefix: @"\ebeginInterfaceDescription "
          andPostfix: @"\n\eendInterfaceDescription\n"
          forEntity: @"desc"];

    [buffer appendString: @"\n\ebeginClassSection\001"];
    [buffer appendFormat: @"%@ (%@)\002\n", [node objectForKey: @"class"],
            [node objectForKey: @"name"]];

    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"conform"];
    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"declared"];
    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"desc"];

    methods = AUTORELEASE (RETAIN ([childNodes objectForKey: @"method"]));
    [childNodes removeObjectForKey: @"method"];
    [self outputNodesOf: childNodes];
    [self outputClassesAndInstanceMethodsFromArray: methods];
    [buffer appendString: @"\n\eendClassSection\n"];
    [self removeLayerFromStack];
}

- (void) _output_ref: (GSXMLNode*) node
{
    [buffer appendFormat: @"\elink\001%@**", [node objectForKey: @"type"]];
    if ([node objectForKey: @"class"])
        {
        [buffer appendFormat: @"%@**", [node objectForKey: @"class"]];
        }
    else if ([[node objectForKey: @"type"] compare: @"method"] == NSOrderedSame)
        {
        [buffer appendFormat: @"%@**", _currentClass];
        }
    [buffer appendFormat: @"%@\002\001", [node objectForKey: @"id"]];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"\002"];
}


//------------------------------------------------------------

- (void) _output_index: (GSXMLNode*) node
{
    NSDictionary *indexMap = nil;
    NSEnumerator *num;
    NSString     *key;
    NSString     *type;
    
    if ([[node objectForKey: @"type"] compare: @"title"] == NSOrderedSame)
        {
        return;
        }

    if ([[node objectForKey: @"scope"] compare: @"project"] == NSOrderedSame)
        {
        type = [node objectForKey: @"type"];

        indexMap = [[projectRefs refs] objectForKey: type];
        }

    num = [[[indexMap allKeys] sortedArrayUsingSelector: @selector(compare:)] objectEnumerator];

    //NOTE THIS IS VERY INNEFICIENT, for every object we reparse the complete .gsdoc file.
    //So a file defining 10 functions, 4 variables and 1 class is parsed 15 times.
    //This couldd be rewritten!
    while ((key = [num nextObject]))
        {
        CREATE_AUTORELEASE_POOL (arp);

        NSString *fileName;
        NSString *insertionText;
        AGSTex *insertionGenerator;
        GSXMLNode *rootNode;
        
        fileName = [theFileManager findFile: [[indexMap objectForKey: key] stringByAppendingPathExtension: @"gsdoc"]];
        
        insertionGenerator = [AGSTex new];

        [insertionGenerator setGlobalRefs: globalRefs];
        [insertionGenerator setProjectRefs: projectRefs];
        
        rootNode = [insertionGenerator rootNodeAfterSetupFromFile: fileName];

        
        insertionText = [insertionGenerator outputDocument: [self _nodeOfType: type
                                                                  defining: key
                                                                  descendendOf: rootNode]];

        [buffer appendFormat: @"\n\005%% including %@\n",
                [[fileName componentsSeparatedByString: @"\n"] objectAtIndex: 0]];
        [buffer appendString: insertionText];

        RELEASE (insertionGenerator);
        RELEASE (arp);
        }
}


@end
