/**
   <title>AGSXlp</title>
   Copyright (C) 2003 Willem Rein Oudshoorn.

   Written by: Willem Rein Oudshoorn <woudshoo@xs4all.nl>
   Created: Februari 2003

   This file will be part of the GNUstep Project

   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   You should have received a copy of the GNU General Public
   License along with this program; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <Foundation/Foundation.h>
#include "AGSXlp.h"


@implementation AGSXlp

-  init
{
    self = [super init];

    _currentClass = @"";
    
    [self setPrefix: @"<b>" andPostfix: @"</b>" forEntity: @"strong"];
    [self setPrefix: @"<em>" andPostfix: @"</em>" forEntity: @"em"];
    [self setPrefix: @"<code>" andPostfix: @"</code>" forEntity: @"code"];
    [self setPrefix: @"<i>" andPostfix: @"</i>" forEntity: @"var"];
    [self setPrefix: @"<br/>" andPostfix: @"\n\n" forEntity: @"desc"];
    [self setPrefix: @"<br/><b>Declared in: </b>" forEntity: @"declared"];
    [self setPrefix: @"<b>" andPostfix: @" </b>" forEntity: @"sel"];
    [self setPrefix: @"\n<br/>\n" forEntity: @"br"];
    [self setPrefix: @"<ul>\n" andPostfix: @"</ul>\n" forEntity: @"list"];
    [self setPrefix: @"<li>" andPostfix: @"</li>" forEntity: @"item"];
    [self setPrefix: @"<ul>\n" andPostfix: @"</ul>\n" forEntity: @"deflist"];
    [self setPrefix: @"<br/><b>" andPostfix: @"</b>" forEntity: @"term"];

    // ignored for the moment

    [self setPrefix: @"" forEntity: @"gsdoc"];
    [self setPrefix: @"" forEntity: @"body"];
    [self setPrefix: @"" forEntity: @"chapter"];
    [self setPrefix: @"" forEntity: @"front"];
    [self setPrefix: @"" forEntity: @"head"];
    [self setPrefix: @"" forEntity: @"contents"];
    [self setPrefix: @"" forEntity: @"conform"];
    [self setPrefix: @"" forEntity: @"back"];
    [self mapName: @"standards" toName: @"nothing"];
    [self mapName: @"front" toName: @"nothing"];
    [self mapName: @"head" toName: @"nothing"];
    [self mapName: @"heading" toName: @"nothing"];
    [self mapName: @"p" toName: @"br"];
    
    return self;
}

- (void) dealloc
{
    RELEASE (_currentClass);
    [super dealloc];
}


- (NSString*) tableOfContents
{
    NSMutableString *output = [NSMutableString string];
    NSEnumerator *num;
    NSDictionary *titleDict;
    NSString *key;
    
    [output appendString: @"<document>\n"];

    titleDict = [[projectRefs refs] objectForKey: @"title"];
    num = [[[titleDict allKeys] sortedArrayUsingSelector: @selector (compare:)] objectEnumerator];
    
    while ((key = [num nextObject]))
        {
        [output appendFormat: @"<chapter src=\"%@.xlp\" name=\"%@\"></chapter>\n",
                key, key];
        }
    
    [output appendString: @"</document>\n"];

    return output;
}

- (void) outputInstanceMethods: (NSArray*) methods
{
    if ([methods count])
        {
        [buffer appendString: @"\n<section name=\"Instance Methods\">\n"];
        [self outputNodeArray: methods];
        [buffer appendString: @"\n</section>"];
        }
}

- (void) outputClassMethods: (NSArray*) methods
{
    if ([methods count])
        {
        [buffer appendString: @"\n<section name=\"Class Methods\">\n"];
        [self outputNodeArray: methods];
        [buffer appendString: @"\n</section>"];
        }
}


/*
  =============================================================================
   The custom output methods for XML entities
  ============================================================================= 
*/

- (void) _output_class: (GSXMLNode*) node
{
    NSMutableDictionary *classNodes = [self collectChildNodesByName: node];

    ASSIGN (_currentClass, [node objectForKey: @"name"]);
    
    [buffer appendFormat: @"\n\n<section name=\"%@\">\n\n", [node objectForKey: @"name"]];
    [buffer appendFormat: @"\n<section name=\"Overview\">"];

    [buffer appendString: @"\n<b>Inherits From:  </b>"];
    if ([node objectForKey: @"super"])
        {
        NSDictionary *mappingToSuper = [[globalRefs refs] objectForKey: @"super"];
        NSString *current = [mappingToSuper objectForKey: [node objectForKey: @"name"]];

        [buffer appendFormat: @"<ref id=\"%@\">%@</ref>", current, current];

        while ((current = [mappingToSuper objectForKey: current]))
            {
            [buffer appendFormat: @": <ref id=\"%@\">%@</ref>", current, current];
            }
        }
    else
        {
        [buffer appendFormat: @"None, is a root class"];
        }

    if ([classNodes objectForKey: @"conform"])
        {
        [buffer appendString: @"\n<br><b>Conforms to: </b>"];
        [self outputNodesOf: classNodes forKeyAndRemoveKey: @"conform"];
        [buffer appendString: @"\n"];
        }

    [self outputNodesOf: classNodes forKeyAndRemoveKey: @"declared"];
    [buffer appendString: @"\n</section>\n\n"];
    [buffer appendString: @"<section name=\"Description\">\n"];
    [self outputNodesOf: classNodes forKeyAndRemoveKey: @"desc"];
    [buffer appendString: @"\n</section>\n\n"];
    
    [self outputClassesAndInstanceMethodsFromArray: [classNodes objectForKey: @"method"]];
    [classNodes removeObjectForKey: @"method"];

    [self outputNodesOf: classNodes]; //SHould not output anything, but just be save.
    [buffer appendString: @"\n</section>\n\n"];
    [buffer appendString: @"\n</section>\n"];

    ASSIGN (_currentClass, @"");
}

- (void) _output_protocol: (GSXMLNode*) node
{
    NSMutableDictionary *childNodes = [self collectChildNodesByName: node];
    NSArray *methods;
    
    ASSIGN (_currentClass, [node objectForKey: @"name"]);
    
    [buffer appendFormat: @"\n\n<chapter name=\"%@ (Protocol)\">\n\n", [node objectForKey: @"name"]];
    [buffer appendFormat: @"\n<section name=\"Overview\">"];

    if ([childNodes objectForKey: @"conform"])
        {
        [buffer appendString: @"\n<br><b>Conforms to: </b>"];
        [self outputNodesOf: childNodes forKeyAndRemoveKey: @"conform"];
        [buffer appendString: @"\n"];
        }

    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"declared"];
    [buffer appendString: @"\n</section>\n\n"];
    [buffer appendString: @"<section name=\"Description\">\n"];
    [self outputNodesOf: childNodes forKeyAndRemoveKey: @"desc"];
    [buffer appendString: @"\n</section>\n\n"];

    methods = AUTORELEASE (RETAIN ([childNodes objectForKey: @"method"]));
    [childNodes removeObjectForKey: @"method"];
    
    [self outputNodesOf: childNodes]; 
    
    [self outputClassesAndInstanceMethodsFromArray: methods];

    [buffer appendString: @"\n</section>\n\n"];
    [buffer appendString: @"\n</chapter>\n"];

    ASSIGN (_currentClass, @"");
}

- (void) _output_ref: (GSXMLNode*) node
{
    [buffer appendFormat: @"<ref id=\"%@**", [node objectForKey: @"type"]];
    if ([node objectForKey: @"class"])
        {
        [buffer appendFormat: @"%@**", [node objectForKey: @"class"]];
        }
    else
        {
        [buffer appendFormat: @"%@**", _currentClass];
        }
    [buffer appendFormat: @"%@\">", [node objectForKey: @"id"]];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"</ref>"];
}

- (void) _output_method: (GSXMLNode*) node
{
    NSMutableDictionary *nodesByName = [self collectChildNodesByName: node];
    NSMutableString *methodName = [NSMutableString string];
    NSMutableString *oldBuffer;

    // Trick to extract the complete method name from the node
    [self  addLayerToStack];
    [self setPrefix: @"" andPostfix: @"" forEntity: @"sel"];
    oldBuffer = buffer;
    buffer = methodName;
    [self outputNodeArray: [nodesByName objectForKey: @"sel"]];    
    buffer = oldBuffer;
    [self removeLayerFromStack];
    // end trick.
    
    [buffer appendFormat: @"\n\n<section name=\"%@\">\n", methodName];
    
    if ([[node objectForKey: @"factory"] boolValue])
        {
        [buffer appendFormat: @"<label id=\"method**%@**+%@\" />", _currentClass, methodName];
        [buffer appendString: @"+ "];
        }
    else
        {
        [buffer appendFormat: @"<label id=\"method**%@**-%@\" />", _currentClass, methodName];
        [buffer appendString: @"- "];
        }
    
    [buffer appendFormat: @"<code>(%@) </code>", [node objectForKey: @"type"]];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"\n</section>\n"];
}

- (void) _output_arg: (GSXMLNode*) node
{
    [buffer appendFormat: @"<code>(%@) </code><i>", [node objectForKey: @"type"]];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @" </i>"];
}

- (void) _output_example: (GSXMLNode*) node
{
    [self addLayerToStack];
    [self mapName: @"text" toName: @"verbatimText"];
    [buffer appendString: @"<listing>"];
    [self outputNodeList: [node firstChild]];
    [buffer appendString: @"</listing>"];
    [self removeLayerFromStack];
}


- (void) _output_text: (GSXMLNode*) node
{
    [buffer appendString: [node content]];
}

- (void) _output_verbatimText: (GSXMLNode*) node
{
    NSString *original = [node content];
    NSArray* components = [original componentsSeparatedByString: @"\n"];

    [buffer appendString: [components componentsJoinedByString: @"<br/>\n"]];
}

- (void) _output_nothing: (GSXMLNode*) node
{
}

- (void) _output_index: (GSXMLNode*) node
{
    NSDictionary *indexMap = nil;
    NSEnumerator *num;
    NSString     *key;
    NSString     *fileName;
    NSString     *type;


    if ([[node objectForKey: @"scope"] compare: @"project"] == NSOrderedSame)
        {
        type = [node objectForKey: @"type"];

        if ([type compare: @"class"] == NSOrderedSame)
            {
            indexMap = [[projectRefs refs] objectForKey: type];
            }
        else if ([type compare: @"protocol"] == NSOrderedSame)
            {
            indexMap = [[projectRefs refs] objectForKey: type];
            }
        }
    
    num = [indexMap keyEnumerator];

    while ((key = [num nextObject]))
        {
        CREATE_AUTORELEASE_POOL (arp);

        NSString *fileName;
        NSString *insertionText;
        AGSXlp *insertionGenerator;
        GSXMLNode *rootNode;
        
        fileName = [[indexMap objectForKey: key] stringByAppendingPathExtension: @"gsdoc"];
        
        insertionGenerator = [AGSXlp new];

        [insertionGenerator setGlobalRefs: globalRefs];
        [insertionGenerator setProjectRefs: projectRefs];
        
        rootNode = [insertionGenerator rootNodeAfterSetupFromFile: fileName];

        insertionText = [insertionGenerator outputDocument: [self _nodeOfType: type
                                                                  defining: key
                                                                  descendendOf: rootNode]];

        [buffer appendString: insertionText];

        RELEASE (insertionGenerator);
        RELEASE (arp);
        }
}

@end
