/** This tool produces gsdoc files from source files.

   <title>Autogsdoc ... a tool to make documentation from source code</title>
   Copyright (C) 2001 Free Software Foundation, Inc.

   Written by:  Richard Frith-Macdonald <richard@brainstorm.co.uk>
   Created: October 2001

   This file is part of the GNUstep Project

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   You should have received a copy of the GNU General Public
   License along with this program; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

<chapter>
  <heading>The autogsdoc tool</heading>
  <section>
    <heading>overview</heading>
    <p>
      The autogsdoc tool is a command-line utility for parsing ObjectiveC
      source code (header files and optionally source files) in order to
      generate documentation covering the public interface of the various
      classes, categories, and protocols in the source.
    </p>
    <p>
      The simple way to use this is to run the command with one or more
      header file names as arguments ... the tool will automatically
      parse corresponding source files in the same directory as the
      headers (or the current directory, or the directory specified
      using the DocumentationDirectory default), and produce gsdoc
      and html files as output.
    </p>
    <p>
      Even without any human assistance, this tool will produce skeleton
      documents listing the methods in the classes found in the source
      files, but more importantly it can take specially formatted comments
      from the source files and insert those comments into the gsdoc output.
    </p>
    <p>
      Any comment beginning with slash and <em>two</em> asterisks rather than
      the common slash and single asterisk, is taken to be gsdoc markup, to
      be use as the description of the class or method following it.  This
      comment text is reformatted and then inserted into the output.<br />
      Where multiple comments are associatd with the same item, they are
      joined together with a line break (&lt;br /&gt;) between each if
      necessary.
    </p>
    <p>
      The tool can easily be used to document programs as well as libraries,
      simply by giving it the name of the source file containing the main()
      function of the program - it takes the special comments from that
      function and handles them specially, inserting them as a section at
      the end of the first chapter of the document (it creates the first
      chapter if necessary).
    </p>
  </section>
  <section>
    <heading>Extra markup</heading>
    <p>
      There are some cases where special extra processing is performed,
      predominantly in the first comment found in the source file,
      from which various chunks of gsdoc markup may be extracted and
      placed into appropriate locations in the output document -
    </p>
    <list>
      <item><strong>AutogsdocSource</strong>
	In any line where <code>AutogsdocSource</code>: is found, the remainder
	of the line is taken as a source file name to be used instead of
	making the assumption that each .h file processed uses a .m file
	of the same name.  You may supply multiple <code>AutogsdocSource</code>:
	lines where a header file declares items which are defined in
	multiple source files.<br />
	If a file name is absolute, it is used just as supplied.<br />
	If on the other hand, it is a relative path, the software looks for
	the source file first relative to the location of the header file,
	and if not found there, relative to the current directory in which
	autogsdoc is running, and finally relative to the directory
	specified by the DocumentationDirectory default.
      </item>
      <item><strong>&lt;abstract&gt;</strong>
	An abstract of the content of the document ... placed in the head
	of the gsdoc output.
      </item>
      <item><strong>&lt;author&gt;</strong>
	A description of the author of the code - may be repeated to handle
	the case where a document has multiple authors.  Placed in the
	head of the gsdoc output.<br />
	As an aid to readability of the source, some special additional
	processing is performed related to the document author -<br />
	Any line of the form '<code>Author</code>: name &lt;email-address&gt;',
	or '<code>By</code>: name &lt;email-address&gt;',
	or '<code>Author</code>: name' or '<code>By</code>: name'
	will be recognised and converted to an <em>author</em> element,
	possibly containing an <em>email</em> element.
      </item>
      <item><strong>&lt;back&gt;</strong>
	Placed in the gsdoc output just before the end of the body of the
	document - intended to be used for appendices, index etc.
      </item>
      <item><strong>&lt;chapter&gt;</strong>
	Placed immediately before any generated class documentation ...
	intended to be used to provide overall description of how the
	code being documented works.<br />Any documentation for the main()
	function of a program is inserted as a section at the end of this
	chapter.
      </item>
      <item><strong>&lt;copy&gt;</strong>
	Copyright of the content of the document ... placed in the head
	of the gsdoc output.<br />
	As an aid to readability of the source, some special additional
	processing is performed -<br />
	Any line of the form 'Copyright (C) text' will be recognised and
	converted to a <em>copy</em> element.
      </item>
      <item><strong>&lt;date&gt;</strong>
	Date of the revision of the document ... placed in the head
	of the gsdoc output.  If this is omitted the tool will try to
	construct a value from the RCS Date tag (if available).
      </item>
      <item><strong>&lt;front&gt;</strong>
	Inserted into the document at the start of the body ... intended
	to provide for introduction or contents pages etc.
      </item>
      <item><strong>&lt;title&gt;</strong>
	Title of the document ... placed in the head of the gsdoc output.
	If this is omitted the tool will generate a (probably poor)
	title of its own - so you should include this markup manually.
      </item>
      <item>
	<strong>NB</strong>This markup may be used within
	class, category, or protocol documentation ... if so, it is
	extracted and wrapped round the rest of the documentation for
	the class as the classes chapter.
	The rest of the class documentation is normally
	inserted at the end of the chapter, but may instead be substituted
	in in place of the &lt;unit /&gt; pseudo-element within the
	&lt;chapter&gt; element.
      </item>
      <item><strong>&lt;version&gt;</strong>
	Version identifier of the document ... placed in the head
	of the gsdoc output.  If this is omitted the tool will try to
	construct a value from the RCS Revision tag (if available).
      </item>
    </list>
  </section>
  <section>
    <heading>Method markup</heading>
    <p>
      In comments being used to provide text for a method description, the
      following markup is removed from the text and handled specially -
    </p>
    <list>
      <item><strong>&lt;init /&gt;</strong>
	The method is marked as being the designated initialiser for the class.
      </item>
      <item><strong>&lt;override-subclass /&gt;</strong>
	The method is marked as being one which subclasses must override
	(eg an abstract method).
      </item>
      <item><strong>&lt;override-never /&gt;</strong>
	The method is marked as being one which subclasses should <em>NOT</em>
	override.
      </item>
      <item><strong>&lt;standards&gt; ... &lt;/standards&gt;</strong>
	The markup is removed from the description and placed <em>after</em>
	it in the gsdoc output - so that the method is described as
	conforming (or not conforming) to the specified standards.
      </item>
    </list>
  </section>
  <section>
    <heading>Automated markup</heading>
    <p>
      Generally, the text in comments is reformatted to standardise and
      indent it nicely ... the reformatting is <em>not</em> performed on
      any text inside an &lt;example&gt; element.<br />
      When the text is reformatted, it is broken into whitespace separated
      'words' which are then subjected to some extra processing ...
    </p>
    <list>
      <item>Certain well known constants such as YES, NO, and nil are
	enclosed in &lt;code&gt; ... &lt;/code&gt; markup.
      </item>
      <item>The names of method arguments within method descriptions are
	enclosed in &lt;var&gt; ... &lt;/var&gt; markup.
      </item>
      <item>Method names (beginning with a plus or minus) are enclosed
	in &lt;ref...&gt; ... &lt;/ref&gt; markup.<br />
	eg. "-init" (without the quotes) would be wrapped in a gsdoc
	reference element to point to the init method of the current
	class or, if only one known class had an init method, it
	would refer to the method of that class.
	<br />Note the fact that the method name must be surrounded by
	whitespace to be recognized (though a comma, fullstop, or semicolon
	at the end of the specifier will act like whitespace).
      </item>
      <item>Method specifiers including class names (beginning and ending with
	square brackets) are enclosed in &lt;ref...&gt; ... &lt;/ref&gt; markup.
	<br />eg. <code>[</code>NSObject-init<code>]</code>,
	will create a reference to the init method of NSObject (either the
	class proper, or any of its categories), while
	<br /><code>[</code>(NSCopying)-copyWithZone:<code>]</code>, creates a
	reference to a method in the NSCopyIng protocol.
	<br />Note that no spaces must appear between the square brackets
	in these specifiers.
	<br />Protocol names are enclosed in round brackets rather than
	the customary angle brackets, because gsdoc is an XML language, and
	XML treats angle brackets specially.
      </item>
      <item>Class names (and also protocol and category names) enclosed
	in square brackets are also cross referenced.
	<br />Protocol names are enclosed in round brackets rather than
	the customary angle brackets, because gsdoc is an XML language, and
	XML treats angle brackets specially.
      </item>
      <item>Function names (ending with '()') other than 'main()' are enclosed
	in &lt;ref...&gt; ... &lt;/ref&gt; markup.<br />
	eg. "NSLogv()" (without the quotes) would be wrapped in a gsdoc
	reference element to point to the documentation of the NSLog function.
	<br />Note the fact that the function name must be surrounded by
	whitespace (though a comma, fullstop, or semicolon at the end
	of the specifier will also act as a whitespace terminator).
      </item>
    </list>
  </section>
  <section>
    <heading>Arguments and Defaults</heading>
    <p>
      The tools accepts certain user defaults (which can of course be
      supplied as command-line arguments as usual) -
    </p>
    <list>
      <item><strong>Clean</strong>
	If this boolean value is set to YES, then rather than generating
	documentation, the tool removes all gsdoc files generated in the
	project, and all html files generated from them (as well as any
	which would be generated from gsdoc files listed explicitly),
	and finally removes the project index file.<br />
	The only exception to this is that template gsdoc files (ie those
	specifield using "-ConstantsTemplate ...", "-FunctionsTemplate ..."
        arguments etc) are not deleted unless the CleanTemplates flag is set.
      </item>
      <item><strong>CleanTemplates</strong>
	This flag specifies whether template gsdoc files are to be removed
	along with other files when the Clean option is specified.
	The default is for them not to be removed ... since these templates
	may have been produced manually and just had data inserted into them.
      </item>
      <item><strong>ConstantsTemplate</strong>
	Specify the name of a template document into which documentation
	about constants should be inserted from all files in the project.<br />
	This is useful if constants in the source code are scattered around many
	files, and you need to group them into one place.<br />
	You are responsible for ensuring that the basic template document
	(into which individual constant documentation is inserted) contains
	all the other information you want, but as a convenience autogsdoc
	will generate a simple template (which you may then edit) for you
	if the file does not exist.
	<br />Insertion takes place immediately before the <em>back</em>
	element (or if that does not exist, immediately before the end
	of the <em>body</em> element) in the template.
      </item>
      <item><strong>Declared</strong>
	Specify where headers are to be documented as being found.<br />
	The actual name produced in the documentation is formed by appending
	the last component of the header file name to the value of this
	default.<br />
	If this default is not specified, the full name of the header file
	(as supplied on the command line), with the HeaderDirectory
	default prepended, is used.<br />
	A typical usage of this might be <code>"-Declared Foundation"</code>
	when generating documentation for the GNUstep base library.  This
	would result in the documentation saying that NSString is declared
	in <code>Foundation/NSString.h</code>
      </item>
      <item><strong>DocumentAllInstanceVariables</strong>
	This flag permits you to generate documentation for all instance
	variables.  Normally, only those explicitly declared 'public' or
	'protected' will be documented.
      </item>
      <item><strong>DocumentationDirectory</strong>
	May be used to specify the directory in which generated
	documentation is to be placed.  If this is not set, output
	is placed in the current directory.  This directory is also
	used as a last resort to locate source files (not headers).
      </item>
      <item><strong>Files</strong>
	Specifies the name of a file containing a list of file names as
	a property list array <em>(name1,name2,...)</em> format.  If this
	is present, filenames in the program argument list are ignored and
	the names in this file are used as the list of names to process.
      </item>
      <item><strong>FunctionsTemplate</strong>
	Specify the name of a template document into which documentation
	about functions should be inserted from all files in the project.<br />
	This is useful if function source code is scattered around many
	files, and you need to group it into one place.<br />
	You are responsible for ensuring that the basic template document
	(into which individual function documentation is inserted) contains
	all the other information you want, but as a convenience autogsdoc
	will generate a simple template (which you may then edit) for you
	if the file does not exist.
	<br />Insertion takes place immediately before the <em>back</em>
	element (or if that does not exist, immediately before the end
	of the <em>body</em> element) in the template.
      </item>
      <item><strong>GenerateHtml</strong>
	May be used to specify if HTML output is to be generated.
	Defaults to YES.
      </item>
      <item><strong>HeaderDirectory</strong>
	May be used to specify the directory to be searched for header files.
	When supplied, this value is prepended to relative header names,
	otherwise the relative header names are interpreted relative to
	the current directory.<br />
	Header files specified as absolute paths are not influenced by this
	default.
      </item>
      <item><strong>IgnoreDependencies</strong>
	A boolean value which may be used to specify that the program should
	ignore file modification times and regenerate files anyway.  Provided
	for use in conjunction with the <code>make</code> system, which is
	expected to manage dependency checking itsself.
      </item>
      <item><strong>LocalProjects</strong>
	This value is used to control the automatic inclusion of local
	external projects into the indexing system for generation of
	cross-references in final document output.<br />
	If set to 'None', then no local project references are done,
	otherwise, the 'Local' GNUstep documentation directory is recursively
	searched for files with a <code>.igsdoc</code> extension, and the
	indexing information from those files is used.<br />
	The value of this string is also used to generate the filenames in
	the cross reference ... if it is an empty string, the path to use
	is assumed to be a file in the same directory where the igsdoc
	file was found, otherwise it is used as a prefix to the name in
	the index.<br />
	NB. Local projects with the same name as the project currently
	being documented will <em>not</em> be included by this mechanism.
	If you wish to include such projects, you must do so explicitly
	using <em>"-Projects ..."</em>
      </item>
      <item><strong>MacrosTemplate</strong>
	Specify the name of a template document into which documentation
	about macros should be inserted from all files in the project.<br />
	This is useful if macro code is scattered around many
	files, and you need to group it into one place.<br />
	You are responsible for ensuring that the basic template document
	(into which individual macro documentation is inserted) contains
	all the other information you want, but as a convenience autogsdoc
	will generate a simple template (which you may then edit) for you
	if the file does not exist.
	<br />Insertion takes place immediately before the <em>back</em>
	element (or if that does not exist, immediately before the end
	of the <em>body</em> element) in the template.
      </item>
      <item><strong>Project</strong>
	May be used to specify the name of this project ... determines the
	name of the index reference file produced as part of the documentation
	to provide information enabling other projects to cross-reference to
	items in this project.
      </item>
      <item><strong>Projects</strong>
	This value may be supplied as a dictionary containing the paths to
	the igsdoc index/reference files used by external projects, along
	with values to be used to map the filenames found in the indexes.<br />
	For example, if a project index (igsdoc) file says that the class
	<code>Foo</code> is found in the file <code>Foo</code>, and the
	path associated with that project index is <code>/usr/doc/proj</code>,
	Then generated html output may reference the class as being in
	<code>/usr/doc/prj/Foo.html</code>
      </item>
      <item><strong>ShowDependencies</strong>
	A boolean value which may be used to specify that the program should
	log which files are being regenerated because of their dependencies
	on other files.
      </item>
      <item><strong>Standards</strong>
	A boolean value used to specify whether the program should insert
	information about standards complience into ythe documentation.
	This should only be used when documenting the GNUstep libraries
	and tools themselves as it assumes that the code being documented
	is part of GNUstep and possibly complies with the OpenStep standard
	or implements MacOS-X compatible methods.
      </item>
      <item><strong>SystemProjects</strong>
	This value is used to control the automatic inclusion of system
	external projects into the indexing system for generation of
	cross-references in final document output.<br />
	If set to 'None', then no system project references are done,
	otherwise, the 'System' GNUstep documentation directory is recursively
	searched for files with a <code>.igsdoc</code> extension, and the
	indexing information from those files is used.<br />
	The value of this string is also used to generate the filenames in
	the cross reference ... if it is an empty string, the path to use
	is assumed to be a file in the same directory where the igsdoc
	file was found, otherwise it is used as a prefix to the name in
	the index.<br />
	NB. System projects with the same name as the project currently
	being documented will <em>not</em> be included by this mechanism.
	If you wish to include such projects, you must do so explicitly
	using <em>"-Projects ..."</em>
      </item>
      <item><strong>TypedefsTemplate</strong>
	Specify the name of a template document into which documentation
	about typedefs should be inserted from all files in the project.<br />
	This is useful if typedef source code is scattered around many
	files, and you need to group it into one place.<br />
	You are responsible for ensuring that the basic template document
	(into which individual typedef documentation is inserted) contains
	all the other information you want, but as a convenience autogsdoc
	will generate a simple template (which you may then edit) for you
	if the file does not exist.
	<br />Insertion takes place immediately before the <em>back</em>
	element (or if that does not exist, immediately before the end
	of the <em>body</em> element) in the template.
      </item>
      <item><strong>Up</strong>
	A string used to supply the name to be used in the 'up' link from
	generated gsdoc documents.  This should normally be the name of a
	file which contains an index of the contents of a project.<br />
	If this is missing or set to an empty string, then no 'up' link
	will be provided in the documents.
      </item>
      <item><strong>VariablesTemplate</strong>
	Specify the name of a template document into which documentation
	about variables should be inserted from all files in the project.<br />
	This is useful if variable source code is scattered around many
	files, and you need to group it into one place.<br />
	You are responsible for ensuring that the basic template document
	(into which individual variable documentation is inserted) contains
	all the other information you want, but as a convenience autogsdoc
	will generate a simple template (which you may then edit) for you
	if the file does not exist.
	<br />Insertion takes place immediately before the <em>back</em>
	element (or if that does not exist, immediately before the end
	of the <em>body</em> element) in the template.
      </item>
      <item><strong>Verbose</strong>
	A boolean used to specify whether you want verbose debug/warning
	output to be produced.
      </item>
      <item><strong>Warn</strong>
	A boolean used to specify whether you want standard warning
	output (eg report of undocumented methods) produced.
      </item>
      <item><strong>WordMap</strong>
	This value is a dictionary used to map identifiers/keywords found
	in the source files  to other words.  Generally you will not have
	to use this, but it is sometimes helpful to avoid the parser being
	confused by the use of C preprocessor macros.  You can effectively
	redefine the macro to something less confusing.<br />
	The value you map the identifier to must be one of -<br />
	Another identifier,<br />
	An empty string - the value is ignored,<br />
	Two slashes ('//') - the rest of the line is ignored.<br />
      </item>
    </list>
  </section>
  <section>
    <heading>Inter-document linkage</heading>
    <p>
      The 'Up' default is used to specify the name of a document which
      should be used as the 'up' link for any other documents used.<br />
      This name must not include a path or extension.<br />
      Generally, the document referred to by this default should be a
      hand-edited gsdoc document which should have a <em>back</em>
      section containing a project index. eg.
    </p>
<example>
  &lt;?xml version="1.0"?&gt;
  &lt;!DOCTYPE gsdoc PUBLIC "-//GNUstep//DTD gsdoc 1.0.0//EN" 
  "http://www.gnustep.org/gsdoc-1_0_0.xml"&gt;
  &lt;gsdoc base="index"&gt;
    &lt;head&gt;
      &lt;title&gt;My project reference&lt;/title&gt;
      &lt;author name="my name"&gt;&lt;/author&gt;
    &lt;/head&gt;
    &lt;body&gt;
      &lt;chapter&gt;
        &lt;heading&gt;My project reference&lt;/heading&gt;
      &lt;/chapter&gt;
      &lt;back&gt;
        &lt;index scope="project" type="title" /&gt;
      &lt;/back&gt;
    &lt;/body&gt;
  &lt;/gsdoc&gt;
</example>
  </section>
</chapter>
<back>
  <index type="title" scope="project" />
  <index type="class" scope="project" />
  <index type="protocol" scope="project" />
</back>
   */

//#include	"../config.h"

#include "autogsdoc.h"
#include "AGSParser.h"
#include "AGSOutput.h"
#include "AGSIndex.h"
#include "AGSHtml.h"
#include "AGSTex.h"
#include "AGSXlp.h"

//#include <Foundation/GNUstep.h>


/* Wim Oudshoorn 14/1/03,

   face it, this is a mess.  The main function
   is now about 1200 lines of code.  That is way too much.

   I want to clean it up.
 */

///

struct 
{
    BOOL generateHtml;
    BOOL generateTeX;
    BOOL generateXLP;
    BOOL ignoreDependencies;
    BOOL showDependencies;
    BOOL clean;
    BOOL cleanTemplates;
    BOOL isVerbose;
    BOOL warn;
    NSString *declared;
    NSString *project;
} options;

struct
{
    NSString     *up;
    NSDictionary *wordMap;
    BOOL          standards;
    BOOL          documentAllInstanceVariables;
} outputOptions;

struct
{
    NSString    *local;
    NSString    *system;
    NSMutableDictionary *normal;
} projects;


@interface FileEnumerator: NSEnumerator
{
    NSEnumerator *_underlyingEnumerator;
    NSString *_basePath;
}

- initWithBasePath: (NSString *) path andEnumerator: (NSEnumerator *) num;

@end


FileManager *theFileManager;



@interface FileManager(Private)

- (void) initializeArrays;
- (void) initializeDirectoriesFromUserDefaults;

@end


@implementation FileManager

- initForProject: (NSString*) project
{
    self = [super init];

    sourceFiles = [[NSMutableArray alloc] init];
    gsdocFiles = [[NSMutableArray alloc] init];
    htmlFiles = [[NSMutableArray alloc] init];

    [self initializeArrays];
    [self initializeDirectoriesFromUserDefaults];

    referenceFile = [documentationDirectory stringByAppendingPathComponent: project];
    referenceFile = [[referenceFile stringByAppendingPathExtension: @"igsdoc"] retain];
    
    return self;
}


- (void) dealloc
{
    [sourceFiles release];
    [gsdocFiles release];
    [htmlFiles release];
    [referenceFile release];

    [documentationDirectory release];
    [headerDirectory release];
    
    [super dealloc];
}

/**
   Initialize the content of the file arrays.
   As a list of input files it will do, listed in order of
   preferences,

   - The content of the file indicate by the userdefault
     with key "Files"

   - Use the list of names given on the command line.

   It will than sort the input files in the three groups:
   source files, gsdoc files, html files.
   The splitting depends on the extension of the file.

   When an error occurs the program will exit immediately.
*/

- (void) initializeArrays
{
    id tmp;
    NSEnumerator *num;
    NSString *fileName;
    
    tmp = [[NSUserDefaults standardUserDefaults]
              stringForKey: @"Files"];

    if (tmp)
        {
        id inputFileArray;
        inputFileArray = [NSArray arrayWithContentsOfFile: tmp];      
        if (inputFileArray == nil)
            {
            NSLog(@"Failed to load files from '%@'", tmp);
            exit(1);
            }
        num = [inputFileArray objectEnumerator];
        }
    else
        {
        num = [[[NSProcessInfo processInfo] arguments] objectEnumerator];
        // Skip fist element because it is the name of the program.
        [num nextObject];
        }

    while ((fileName = [num nextObject]))
        {
        if ([fileName hasPrefix: @"-"])
            {
            // Is argument, skip value of argument as well.
            [num nextObject];
            continue;
            }
        if ([fileName hasSuffix: @".h"]
            || [fileName hasSuffix: @".m"])
            {
            [sourceFiles addObject: fileName];
            continue;
            }
        if ([fileName hasSuffix: @".gsdoc"])
            {
            [gsdocFiles addObject: fileName];
            continue;
            }
        if ([fileName hasSuffix: @".html"])
            {
            [htmlFiles addObject: fileName];
            continue;
            }
        NSLog (@"Filename: `%@' is given as input file,"
               @" but is of uknown type.  Will ignore this file.",
               fileName);
        }
}


/**
   Initialize the instance variables that hold the Headers directory and
   the resulting documentation directory.

   The two directory paths are fetched from the standardUserDefaults,
   so they can be specified on the command line.  The keys in the
   standardUserDefaults are:

   - "HeaderDirectory"
   - "DocumentationDirectory"
*/
- (void) initializeDirectoriesFromUserDefaults
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    NSString        *tmp;

    tmp = [defaults stringForKey: @"HeaderDirectory"];
    if (tmp)
        {
        ASSIGN (headerDirectory, tmp);
        }
    else
        {
        ASSIGN (headerDirectory, @"");
        }
    
    tmp = [defaults stringForKey: @"DocumentationDirectory"];
    if (tmp)
        {
        ASSIGN (documentationDirectory, tmp);
        }
    else
        {
        ASSIGN (documentationDirectory, @"");
        }
}


// public method, documented in the header
- (NSDictionary*) contentsOfReferenceFile
{
    NSFileManager *fMgr = [NSFileManager defaultManager];
    NSDictionary  *result = nil;
    
    if ([fMgr isReadableFileAtPath: referenceFile])
        {
        result = [NSDictionary dictionaryWithContentsOfFile: referenceFile];

        if (result == nil)
            {
            NSLog (@"Unable to read project file `%@', will ignore it",
                   referenceFile);
            }
        }
    return result;
}

// public method, documented in the header
- (BOOL) writeContentsOfReferenceFile: (NSDictionary*) content
{
    if (![content writeToFile: referenceFile atomically: YES])
        {
        NSLog (@"Can not write reference file: `%@'", referenceFile);
        return NO;
        }
    else
        {
        return YES;
        }
}

// public method, documented in the header
- (NSDate*) timeStampOfReferenceFile
{
    NSFileManager *fMgr = [NSFileManager defaultManager];
    NSDate *result;

    if ([fMgr isReadableFileAtPath: referenceFile])
        {
        NSDictionary *fAttr = [fMgr fileAttributesAtPath: referenceFile
                                    traverseLink: YES];

        result = [fAttr fileModificationDate];
        }
    else
        {
        result = [NSDate distantPast];
        }

    return result;
}

//public mehtod, documented in the header
- (void) deleteReferenceFile
{
    NSFileManager *fMgr = [NSFileManager defaultManager];

    if ([fMgr fileExistsAtPath: referenceFile])
        {
        if (![fMgr removeFileAtPath: referenceFile handler: nil])
            {
            NSLog (@"Failed to remove reference file: `%@'", referenceFile);
            }
        }
}

//public method, documented in the header
- (NSString*) fileNameForTemplate: (NSString*) templateName
{
    NSFileManager *fMgr = [NSFileManager defaultManager];
    NSString *key       = [templateName stringByAppendingString: @"Template"];
    NSString *result    = [[NSUserDefaults standardUserDefaults] stringForKey: key];

    result = [result stringByAppendingPathExtension: @"gsdoc"];

    if (result && ![result isAbsolutePath])
        {
        result = [documentationDirectory stringByAppendingPathComponent: result];
        }

    if (![fMgr isReadableFileAtPath: result])
        {
        result = nil;
        }
    
    return result;
}


- (NSDictionary*) templateFileNamesFor: (NSArray*) templateNames
{
    NSMutableDictionary *result = [NSMutableDictionary dictionary];
    NSEnumerator *num = [templateNames objectEnumerator];
    id key;
    id value;

    while ((key = [num nextObject]))
        {
        value = [self fileNameForTemplate: key];
        if (value)
            {
            [result setObject: value forKey: key];
            }
        }

    return result;
}

//public method, documented in the header
- (NSEnumerator*) gsdocFileEnumerator
{
    return AUTORELEASE ([[FileEnumerator alloc] initWithBasePath: documentationDirectory
                            andEnumerator: [gsdocFiles objectEnumerator]]);
}

- (NSEnumerator*) htmlFileEnumerator
{
    return [htmlFiles objectEnumerator];
}

- (NSEnumerator*) sourceFileEnumerator
{
    if ([sourceFiles count] == 0)              // CHECKME, base might already do this optimization!
        {
        return nil; 
        }
    else
        {
        return [sourceFiles objectEnumerator];
        }
}

//public method, documented in the header
- (NSString*) htmlFileForGsdocFile: (NSString*) gsdocFile
{
    NSString *result = [gsdocFile lastPathComponent];
    result = [result stringByDeletingPathExtension];
    result = [result stringByAppendingPathExtension: @"html"];
    result = [documentationDirectory stringByAppendingPathComponent: result];

    return result;
}
//public method, documented in the header
- (NSString*) texFileForGsdocFile: (NSString*) gsdocFile
{
    NSString *result = [gsdocFile lastPathComponent];
    result = [result stringByDeletingPathExtension];
    result = [result stringByAppendingPathExtension: @"tex"];
    result = [documentationDirectory stringByAppendingPathComponent: result];

    return result;
}

- (NSString*) xlpFileForGsdocFile: (NSString*) gsdocFile
{
    NSString *result = [gsdocFile lastPathComponent];
    result = [result stringByDeletingPathExtension];
    result = [result stringByAppendingPathExtension: @"xlp"];
    result = [documentationDirectory stringByAppendingPathComponent: result];

    return result;
}

//public method, documented in the header
- (NSString*) gsdocFileForSourceFile: (NSString*) sourceFile
{
    NSString *result = [sourceFile stringByDeletingPathExtension];
    result = [result lastPathComponent];
    result = [documentationDirectory stringByAppendingPathComponent: result];
    result = [result stringByAppendingPathExtension: @"gsdoc"];

    return result;
}

//public method, documented in the header
- (NSString*) fixSourceFilePath: (NSString*) sourceFile
{
    if (![sourceFile isAbsolutePath]
        && [headerDirectory length] > 0
        && [[sourceFile pathExtension] isEqual: @"h"])
        {
            return [headerDirectory stringByAppendingPathComponent: sourceFile];
        }

    return sourceFile;
}

//public method, documented in the header
- (NSString*) fileInDocumentationDirectory: (NSString*) fileName
{
    if (![fileName isAbsolutePath])
        {
        return [documentationDirectory stringByAppendingPathComponent: fileName]; //???? CHECKME this seems a bug
        }
    else
        {
        return fileName;
        }
}

- (NSString*) findFile: (NSString*) fileName
{
    if (fileName && ![fileName isAbsolutePath]
        && ![[NSFileManager defaultManager] isReadableFileAtPath: fileName])
        {
        NSString *newFileName;

        newFileName = [documentationDirectory stringByAppendingPathComponent: fileName];
        if ([[NSFileManager defaultManager] isReadableFileAtPath: newFileName])
            {
            fileName = newFileName;
            }
        }
    return fileName;
}

- (NSString*) documentationDirectory
{
    return documentationDirectory;
}

- (void) addFileToGsdocs: (NSString*) fileName
{
    NSFileManager *fMgr = [NSFileManager defaultManager];
    NSString *tmp;
    
    fileName = [fileName lastPathComponent];
    fileName = [fileName stringByDeletingPathExtension];
    fileName = [fileName stringByAppendingPathExtension: @"gsdoc"];
    
    // Note this logic tries to follow the logic in the original.
    // Although I am not quite sure if this logic is what is wanted.

    tmp = [headerDirectory stringByAppendingPathComponent: fileName];

    if ([fMgr isReadableFileAtPath: tmp])
        {
        fileName = tmp;
        }
    
    if (![gsdocFiles containsObject: fileName])
        {
        [gsdocFiles addObject: fileName];
        }
}


@end


@implementation FileEnumerator

- initWithBasePath: (NSString*) path andEnumerator: (NSEnumerator*) num
{
    self = [super init];
    
    _underlyingEnumerator = [num retain];
    _basePath = [path copy];

    return self;
}

- (void) dealloc
{
    RELEASE (_underlyingEnumerator);
    RELEASE (_basePath);

    [super dealloc];
}

- (NSArray*) allObjects
{
    return nil;//FIXME
}

- (id) nextObject
{
    NSString *fileName = [_underlyingEnumerator nextObject];

    if (fileName && ![fileName isAbsolutePath]
        && ![[NSFileManager defaultManager] isReadableFileAtPath: fileName])
        {
        NSString *newFileName;

        newFileName = [_basePath stringByAppendingPathComponent: fileName];
        if ([[NSFileManager defaultManager] isReadableFileAtPath: newFileName])
            {
            fileName = newFileName;
            }
        }

    return fileName;
}

@end


struct
{
    AGSIndex       *projectRefs;
    AGSIndex       *globalRefs;
} objects;


void setOptionsToDefaultValues (void)
{
    options.generateHtml = YES;
    options.generateTeX = NO;
    options.generateXLP = NO;
    options.ignoreDependencies = NO;
    options.showDependencies = NO;
    options.isVerbose = NO;
    options.clean = NO;
    options.cleanTemplates = NO;
    options.warn = NO;
    options.declared = nil;
    options.project = @"Untitled";
}

void setOutputOptionsToDefaultValues (void)
{
    outputOptions.up = nil;
    outputOptions.wordMap = nil; 
    outputOptions.standards = NO; 
    outputOptions.documentAllInstanceVariables = NO ;
}

void setProjectsToDefaultValues (void)
{
    projects.local = @"";
    projects.system = @"";
    projects.normal = nil;
}


void initializeProjectsFromUserDefaults (void)
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    NSString *tmp;

    tmp = [defaults stringForKey: @"LocalProjects"];
    if (tmp)
        {
        projects.local = [tmp copy];        
        }

    tmp = [defaults stringForKey: @"SystemProjects"];
    if (tmp)
        {
        projects.system = [tmp copy];
        }

    projects.normal = [[[defaults dictionaryForKey: @"Projects"] mutableCopy] autorelease]; 
}

void initializeOutputOptionsFromUserDefaults (void)
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    
    outputOptions.up = [defaults stringForKey: @"Up"];
    outputOptions.wordMap = [defaults dictionaryForKey: @"WordMap"];
    outputOptions.standards = [defaults boolForKey: @"Standards"];
    outputOptions.documentAllInstanceVariables = [defaults boolForKey: @"Documentallinstancevariables"];
}

        


void initializeOptionsFromUserDefaults (void)
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];

    // WO CHECKME:  why is this put in the defaults?
    [defaults registerDefaults: [NSDictionary dictionaryWithObjectsAndKeys:
                                                  @"Untitled", @"Project",
                                              nil, nil]];
    options.generateTeX        = [defaults boolForKey: @"TeX"];
    options.generateXLP        = [defaults boolForKey: @"XLP"];
    options.isVerbose          = [defaults boolForKey: @"Verbose"];
    options.warn               = [defaults boolForKey: @"Warn"];
    options.ignoreDependencies = [defaults boolForKey: @"IgnoreDependencies"];
    options.showDependencies   = [defaults boolForKey: @"ShowDependencies"];
    options.clean              = [defaults boolForKey: @"Clean"];
    options.cleanTemplates     = [defaults boolForKey: @"CleanTemplates"];
    
    if (options.ignoreDependencies && options.showDependencies)
        {
        options.showDependencies = NO;
        NSLog (@"ShowDependencies(YES) used with IgnoreDependencies(YES),  showDependencies reset to (NO)");
        }

    if ([defaults objectForKey: @"GenerateHtml"])
        {
        options.generateHtml = [defaults boolForKey: @"GenerateHtml"];
        }

    options.declared = [defaults stringForKey: @"Declared"];
    options.project  = [defaults stringForKey: @"Project"];
}


void clean ()
{
    NSDictionary	*output;
    NSEnumerator    *fileEnumerator;
    NSArray		*outputNames;
    NSMutableSet	*allPaths;
    NSSet		*preserve = nil;
    NSString		*path;
    NSFileManager       *mgr;
    NSDictionary *templatePaths;
        
    NSArray	*keys = [NSArray arrayWithObjects:
                                 @"Constants",
                             @"Functions",
                             @"Macros",
                             @"Typedefs",
                             @"Variables",
                             nil];
    

    templatePaths = [theFileManager templateFileNamesFor: keys];
    
/*
 * Build a set of all template files.
 */
    mgr = [NSFileManager defaultManager];
//    templates = AUTORELEASE([NSMutableSet new]);
    
    /*
     * Unless we are supposed to clean templates, we preserve any
     * template gsdoc files, but remove any generated content.
     */
    if (!options.cleanTemplates)
        {
        preserve = [NSSet setWithArray: [templatePaths allValues]];
        fileEnumerator = [templatePaths objectEnumerator];

        while ((path = [fileEnumerator nextObject]))
            {
            NSMutableString	*fileContent;
            NSEnumerator    *keyEnumerator = [keys objectEnumerator];
            NSString		*key;
            unsigned		length;
                
            fileContent = [[NSMutableString alloc] initWithContentsOfFile: path];
            if (!fileContent)
                {
                NSLog(@"Cleaning ... failed to read '%@'", path);
                continue;
                }
            length = [fileContent length];
            while ((key = [keyEnumerator nextObject]))
                {
                NSString	*startString;
                NSString	*endString;
                NSRange	startLocation;
                NSRange	endLocation;
                    
                startString = [NSString stringWithFormat: @"<!--Start%@-->", key];
                startLocation = [fileContent rangeOfString: startString];
                endString = [NSString stringWithFormat: @"<!--End%@-->", key];
                endLocation = [fileContent rangeOfString: endString];
                
                if (startLocation.length > 0 && endLocation.length > 0
                    && endLocation.location > startLocation.location)
                    {
                    NSRange	r;
                        
                    r.location = startLocation.location;
                    r.length = NSMaxRange (endLocation) - r.location;
                    [fileContent replaceCharactersInRange: r withString: @""];
                    }
                }
            if ([fileContent length] != length)
                {
                if (![fileContent writeToFile: path atomically: YES])
                    {
                    NSLog(@"Cleaning ... failed to write '%@'", path);
                    }
                }
            }
        }
    
    /*
     * Build a list of all generated gsdoc files, then remove them
     * and their corresponding html documents.
     */
    output = [[objects.projectRefs refs] objectForKey: @"output"];
    fileEnumerator = [output objectEnumerator];
    allPaths = [[NSMutableSet alloc] initWithArray: [templatePaths allValues]];   // CHECKME, why alloc init?

    while ((outputNames = [fileEnumerator nextObject]))
        {
        [allPaths addObjectsFromArray: outputNames];
        }

    fileEnumerator = [allPaths objectEnumerator];
    while ((path = [fileEnumerator nextObject]))
        {
        /*
         * Delete any gsdoc files which are not in the preserve set.
         */
        if (![preserve member: path])
            {
            if ([mgr fileExistsAtPath: path])
                {
                if ([mgr removeFileAtPath: path handler: nil] == NO)
                    {
                    NSLog(@"Cleaning ... failed to remove %@", path);
                    }
                }
            }
        path = [path stringByDeletingPathExtension];
        path = [path stringByAppendingPathExtension: @"html"];
        if ([mgr fileExistsAtPath: path])
            {
            if (![mgr removeFileAtPath: path handler: nil])
                {
                NSLog(@"Cleaning ... failed to remove %@", path);
                }
            }
        }
    RELEASE(allPaths);
    
    /*
     * Remove the project index file.
     */
    [theFileManager deleteReferenceFile];
    
    /*
     * Remove any HTML documents resulting from gsdoc files which
     * were specified on the command line rather than generated.
     */
    fileEnumerator = [theFileManager gsdocFileEnumerator];

    while ((path = [fileEnumerator nextObject]))
        {
        path = [theFileManager htmlFileForGsdocFile: path];
        if ([mgr fileExistsAtPath: path])
            {
            if (![mgr removeFileAtPath: path handler: nil])
                {
                NSLog(@"Cleaning ... failed to remove %@", path);
                }
            }
        }
}



void processSourceFiles (void)
{
    NSEnumerator *sourceFileEnumerator = [theFileManager sourceFileEnumerator];
    
    if (sourceFileEnumerator)
        {
        CREATE_AUTORELEASE_POOL (arp);

        AGSParser		*parser;
        AGSOutput		*output;
        NSFileManager   *mgr;
        NSString        *sourceFile;
        mgr    = [NSFileManager defaultManager];

        output = [AGSOutput new];

        parser = [AGSParser new];
        [parser setWordMap: outputOptions.wordMap];
        [parser setGenerateStandards: outputOptions.standards];
        [parser setDocumentAllInstanceVariables: outputOptions.documentAllInstanceVariables];
      
        while ((sourceFile = [sourceFileEnumerator nextObject]))
            {
            CREATE_AUTORELEASE_POOL (inner_pool);
            
            NSString		*gsdocFile;
            NSString		*baseName;   // Source file without extension.
            NSDate		*sDate = nil;
            NSDate		*gDate = nil;
          
            /*
             * Note the name of the header file without path or extension.
             * This will be used to generate the output file.
             */
            baseName = [sourceFile stringByDeletingPathExtension];
            baseName = [baseName lastPathComponent];

            gsdocFile = [theFileManager gsdocFileForSourceFile: sourceFile];

            /*
             * Ensure that header file name is set up using the
             * header directory specified unless it is absolute.
             */
            sourceFile = [theFileManager fixSourceFilePath: sourceFile];
          
            if (options.ignoreDependencies == NO)
                {
                NSMutableArray	*dependOnFiles;
                NSMutableArray  *outputFiles;
                NSString *file;
                NSString *outputFile;
                NSEnumerator *dependOnEnumerator;
                NSEnumerator *outputEnumerator;
                /*
                 * Ask existing project info (.gsdoc file) for dependency
                 * information.  Then check the dates on the source files
                 * and the header file.
                 */
                dependOnFiles = [objects.projectRefs sourcesForHeader: sourceFile];
                [dependOnFiles insertObject: sourceFile atIndex: 0];
                dependOnEnumerator = [dependOnFiles objectEnumerator];

                // --- determine the latest modification time of all the source
                // --- files.
                while ((file = [dependOnEnumerator nextObject]))
                    {
                    NSDictionary		*attrs;
                    NSDate	*d;
              
                    attrs = [mgr fileAttributesAtPath: file traverseLink: YES];
                    d = [attrs fileModificationDate];
                    if (sDate == nil || [d earlierDate: sDate] == sDate)
                        {
                        sDate = d;
                        }
                    }
                if (options.isVerbose)
                    {
                    NSLog(@"Saved sources for %@ are %@ ... %@", sourceFile, dependOnFiles, sDate);
                    }

                /*
                 * Ask existing project info (.gsdoc file) for dependency
                 * information.  Then check the dates on the output files.
                 * If none are set, assume the default.
                 */
                // -- determine the earliest modification time for all the
                // -- outputes generated.
                outputFiles = [objects.projectRefs outputsForHeader: sourceFile];
                if ([outputFiles count] == 0)
                    {
                    [outputFiles insertObject: gsdocFile atIndex: 0];
                    }

                outputEnumerator = [outputFiles objectEnumerator];

                while ((outputFile = [outputEnumerator nextObject]))
                    {
                    NSDictionary *attrs;
                    NSDate *d;
                    
                    attrs = [mgr fileAttributesAtPath: outputFile traverseLink: YES];
                    d = [attrs fileModificationDate];
                    if (gDate == nil || [d laterDate: gDate] == gDate)
                        {
                        gDate = d;
                        AUTORELEASE(RETAIN(gDate));
                        }
                    }
                if (options.isVerbose)
                    {
                    NSLog(@"Saved outputs for %@ are %@ ... %@", sourceFile, outputFiles, gDate);
                    }
                }
      
            if (gDate == nil || [sDate earlierDate: gDate] == gDate)
                {
                NSArray	*modified;
                NSMutableArray *a;
                int j;
                
                if (options.showDependencies)
                    {
                    NSLog(@"%@: source %@, gsdoc %@ ==> regenerate",
                          baseName, sDate, gDate);
                    }
                [parser reset];
          
                /*
                 * Try to parse header to see what needs documenting.
                 * If the header given was actually a .m file, this will
                 * parse that file for declarations rather than definitions.
                 */
                if ([mgr isReadableFileAtPath: sourceFile] == NO)
                    {
                    NSLog(@"No readable header at '%@' ... skipping", sourceFile);
                    continue;
                    }
                if (options.declared)
                    {
                    [parser setDeclared:
                                [options.declared stringByAppendingPathComponent:
                                            [sourceFile lastPathComponent]]];
                    }
                [parser parseFile: sourceFile isSource: NO];

                /*
                 * Record dependency information.
                 */
                a = [parser outputs];
                if ([a count] > 0)
                    {
                    int j;
                    /*
                     * Adjust the location of the output files to be in the
                     * documentation directory.
                     */
                    for (j = 0; j < [a count]; j++)
                        {
                        NSString	*s = [a objectAtIndex: j];
                  
                        if ([s isAbsolutePath] == NO)
                            {
                            s = [theFileManager fileInDocumentationDirectory: s];
                            [a replaceObjectAtIndex: j withObject: s];
                            }
                        }
                    if (options.isVerbose)
                        {
                        NSLog(@"Computed outputs for %@ are %@", sourceFile, a);
                        }
                    [objects.projectRefs setOutputs: a forHeader: sourceFile];
                    }
                a = [parser sources];
                if ([a count] > 0)
                    {
                    [objects.projectRefs setSources: a forHeader: sourceFile];
                    }
                if (options.isVerbose)
                    {
                    NSLog(@"Computed sources for %@ are %@", sourceFile, a);
                    }
          
                for (j = 0; j < [a count]; j++)
                    {
                    NSString	*sfile = [a objectAtIndex: j];
              
                    /*
                     * If we can read a source file, parse it for any
                     * additional information on items found in the header.
                     */
                    if ([mgr isReadableFileAtPath: sfile] == YES)
                        {
                        [parser parseFile: sfile isSource: YES];
                        }
                    else
                        {
                        NSLog(@"No readable source at '%@' ... ignored", sfile);
                        }
                    }
          
                /*
                 * Set up linkage for this file.
                 */
                [[parser info] setObject: baseName forKey: @"base"];
                [[parser info] setObject: [theFileManager documentationDirectory] forKey: @"directory"];
          
                /*
                 * Only produce linkage if the up link is not empty.
                 * Don't add an up link if this *is* the up link document.
                 */
                if ([outputOptions.up length] > 0 && [outputOptions.up isEqual: baseName] == NO)
                    {
                    [[parser info] setObject: outputOptions.up forKey: @"up"];
                    }

                modified = [output output: [parser info]];
                if (modified == nil)
                    {
                    NSLog(@"Sorry unable to write %@", gsdocFile);
                    }
                else
                    {
                    unsigned	c = [modified count];
              
                    while (c-- > 0)
                        {
                        NSString	*f;
                  
                        f = [[modified objectAtIndex: c] lastPathComponent];

                        [theFileManager addFileToGsdocs: f];
                        }
                    }
                }
            else
                {
                /*
                 * Add the .h file to the list of those to process.
                 */
                [theFileManager addFileToGsdocs: sourceFile];
                }
            DESTROY (inner_pool);
            }

        DESTROY(parser);
        DESTROY(output);

        DESTROY(arp);
        }
}


//=============================================================================================

void processGSDocFiles (NSDictionary *originalIndex, NSDate *rDate)
{
    CREATE_AUTORELEASE_POOL(arp);

    NSEnumerator *gsdocEnumerator = [theFileManager gsdocFileEnumerator];
    
    if (gsdocEnumerator)
        {
        NSDictionary	*projectIndex;
        NSFileManager   *mgr;
        NSString        *gsdocfile;
        
        mgr = [NSFileManager defaultManager];
        
        while ((gsdocfile = [gsdocEnumerator nextObject]))
            {
            CREATE_AUTORELEASE_POOL (inner_pool);
            
            NSDictionary	*attrs;
            NSDate	        *gDate = nil;

            if (options.ignoreDependencies == NO)
                {
                attrs = [mgr fileAttributesAtPath: gsdocfile traverseLink: YES];
                gDate = [attrs fileModificationDate];
                }
          
            
            /*
             * Now we try to process the gsdoc data to make index info
             * unless the project index is already more up to date than
             * this file (or the gsdoc file does not exist of course).
             */
            if (options.ignoreDependencies || [gDate earlierDate: rDate] == rDate)
                {
                GSXMLNode	*root;
                GSXMLParser	*parser;
                AGSIndex	*localRefs;

                if (options.showDependencies == YES)
                    {
                    NSLog(@"%@: gsdoc %@, index %@ ==> regenerate",
                          gsdocfile, gDate, rDate);
                    }

                parser = [GSXMLParser parserWithContentsOfFile: gsdocfile];
                [parser substituteEntities: NO];
                [parser doValidityChecking: YES];
                [parser keepBlanks: NO];
                if ([parser parse] == NO)
                    {
                    NSLog(@"1-WARNING %@ is not a valid document", gsdocfile);
                    }
                root = [[parser document] root];
                if (![[root name] isEqualToString: @"gsdoc"])
                    {
                    NSLog(@"not a gsdoc document - because name node is %@",
                          [root name]);
                    exit (1);  // WO FIXME, exit more gracefully
                    }
                  
                localRefs = AUTORELEASE([AGSIndex new]);
                [localRefs makeRefs: root];
                  
                /*
                 * accumulate index info in project references
                 */
                [objects.projectRefs mergeRefs: [localRefs refs] override: NO];
                }
            DESTROY (inner_pool);
            }

        /*
         * Save project references if they have been modified.
         */
        projectIndex = [objects.projectRefs refs];
        if (projectIndex != nil && [originalIndex isEqual: projectIndex] == NO)
            {
            [theFileManager writeContentsOfReferenceFile: projectIndex];
            }
        }
    DESTROY(arp);
}



void createCrossReferences (void)
{
    CREATE_AUTORELEASE_POOL (pool);
    NSFileManager *mgr;
    NSString *refsName;


    mgr = [NSFileManager defaultManager];
    refsName = [options.project stringByAppendingPathExtension: @"igsdoc"];
    
    setProjectsToDefaultValues ();
    initializeProjectsFromUserDefaults ();
      
    /*
     * Merge any external project references into the
     * main cross reference index.
     */
    if ([projects.system caseInsensitiveCompare: @"None"] != NSOrderedSame)
        {
        NSString	*base = [NSSearchPathForDirectoriesInDomains(
                                 NSDocumentationDirectory, NSSystemDomainMask, NO) lastObject];
          
        base = [base stringByStandardizingPath];
        if (base != nil)
            {
            NSDirectoryEnumerator *enumerator = [mgr enumeratorAtPath: base];
            NSString		*file;
              
            if ([projects.system isEqual: @""] == YES)
                {
                projects.system = base;	// Absolute path
                }
            while ((file = [enumerator nextObject]) != nil)
                {
                NSString	*ext = [file pathExtension];
                  
                if ([ext isEqualToString: @"igsdoc"] == YES
                    && [[file lastPathComponent] isEqual: refsName] == NO)
                    {
                    NSString	*key;
                    NSString	*val;

                    if (projects.normal == nil)
                        {
                        projects.normal = [NSMutableDictionary dictionary];
                        }
                    key = [base stringByAppendingPathComponent: file];
                    val = [file stringByDeletingLastPathComponent];
                    val
                        = [projects.system stringByAppendingPathComponent: val];
                    [projects.normal setObject: val forKey: key];
                    }
                }
            }
        }

    if ([projects.local caseInsensitiveCompare: @"None"] != NSOrderedSame)
        {
        NSString	*base = [NSSearchPathForDirectoriesInDomains(
                                 NSDocumentationDirectory, NSLocalDomainMask, NO) lastObject];

        base = [base stringByStandardizingPath];
        if (base != nil)
            {
            NSDirectoryEnumerator	*enumerator;
            NSString			*file;

            enumerator = [mgr enumeratorAtPath: base];
            if ([projects.local isEqual: @""] == YES)
                {
                projects.local = base;	// Absolute path
                }
            while ((file = [enumerator nextObject]) != nil)
                {
                NSString	*ext = [file pathExtension];
		  

                if ([ext isEqualToString: @"igsdoc"] == YES
                    && [[file lastPathComponent] isEqual: refsName] == NO)
                    {
                    NSString	*key;
                    NSString	*val;

                    if (projects.normal == nil)
                        {
                        projects.normal = [NSMutableDictionary dictionary];
                        }
                    key = [base stringByAppendingPathComponent: file];
                    val = [file stringByDeletingLastPathComponent];
                    val = [projects.local stringByAppendingPathComponent: val];
                    [projects.normal setObject: val forKey: key];
                    }
                }
            }
        }

    if (projects.normal != nil)
        {
        NSEnumerator	*e = [projects.normal keyEnumerator];
        NSString	*k;

        while ((k = [e nextObject]) != nil)
            {
            if ([mgr isReadableFileAtPath: k] == NO)
                {
                NSLog(@"Unable to read project file '%@'", k);
                }
            else
                {
                NSDictionary	*dict;

                dict = [[NSDictionary alloc] initWithContentsOfFile: k];

                if (dict == nil)
                    {
                    NSLog(@"Unable to read project file '%@'", k);
                    }
                else
                    {
                    AGSIndex		*tmp;
                    NSString		*p;

                    tmp = [AGSIndex new];
                    [tmp mergeRefs: dict override: NO];
                    RELEASE(dict);
                    /*
                     * Adjust path to external project files ...
                     */
                    p = [projects.normal objectForKey: k];
                    if ([p isEqual: @""] == YES)
                        {
                        p = [k stringByDeletingLastPathComponent];
                        }
                    [tmp setDirectory: p];
                    [objects.globalRefs mergeRefs: [tmp refs] override: YES];
                    RELEASE(tmp);
                    }
                }
            }
        }

    /*
     * Accumulate project index info into global index
     */
    [objects.globalRefs mergeRefs: [objects.projectRefs refs] override: YES];

#if GS_WITH_GC == 0
    RELEASE(pool);
#endif
}


/**
   Returns YES if the first file is newer than the second
 */
BOOL firstNewerThanSecond (NSString *fileA, NSString *fileB)
{
    NSDate *dateA;
    NSDate *dateB;
    NSDictionary *attributes;
    NSFileManager *fMgr = [NSFileManager defaultManager];
    
    attributes = [fMgr fileAttributesAtPath: fileA traverseLink: YES];
    dateA = [attributes fileModificationDate];
    attributes = [fMgr fileAttributesAtPath: fileB traverseLink: YES];
    dateB = [attributes fileModificationDate];

    if (dateB)
        {
        return [dateA earlierDate: dateB] == dateB;
        }
    else
        { // if dateB == nil means fileB does not exist, so assume B older
        return YES;
        }
}




void createHtmlOutput (void)
{
    CREATE_AUTORELEASE_POOL (arp);
    
    NSEnumerator *gsdocEnumerator = [theFileManager gsdocFileEnumerator];
    NSString	 *gsdocfile;

    while ((gsdocfile = [gsdocEnumerator nextObject]))
        {
        CREATE_AUTORELEASE_POOL (inner_pool);
        NSString	*htmlfile;
        NSString	*generated;

        htmlfile = [theFileManager htmlFileForGsdocFile: gsdocfile];

        if (options.ignoreDependencies || firstNewerThanSecond (gsdocfile, htmlfile))
            {
            GSXMLNode	*root;
            GSXMLParser	*parser;
            AGSIndex	*localRefs;
            AGSHtml	*html;

            if (options.showDependencies == YES)
                {
                NSLog(@"%@ -> %@  ==> regenerate", gsdocfile, htmlfile);
                }

            // ---- 1 Parse gsdoc file ---------------------------------------------
            parser = [GSXMLParser parserWithContentsOfFile: gsdocfile];
            [parser substituteEntities: NO];
            [parser doValidityChecking: YES];
            [parser keepBlanks: NO];
            if ([parser parse] == NO)
                {
                NSLog(@"2-WARNING %@ is not a valid document", gsdocfile);
                }
            root = [[parser document] root];
            if (![[root name] isEqualToString: @"gsdoc"])
                {
                NSLog(@"not a gsdoc document - because name node is %@",
                      [root name]);
//                exit (1);  // FIXME better exit
                continue;  //FIXME, figure out why this is reached in my version and not in the original
                }

            // ---- 2 Create references -----------------------------------------
            localRefs = [AGSIndex new];
            [localRefs makeRefs: root];


            // ---- 3 Write html file -------------------------------------------
            html = [AGSHtml new];
            [html setGlobalRefs: objects.globalRefs];
            [html setProjectRefs: objects.projectRefs];
            [html setLocalRefs: localRefs];
            generated = [html outputDocument: root];
            if ([generated writeToFile: htmlfile atomically: YES] == NO)
                {
                NSLog(@"Sorry unable to write %@", htmlfile);
                }
            RELEASE (html);
            RELEASE (localRefs);
            }
        DESTROY (inner_pool);
        }
    DESTROY (arp);
}


void createTeXOutput (void)
{
    CREATE_AUTORELEASE_POOL (arp);
    
    NSEnumerator *gsdocEnumerator = [theFileManager gsdocFileEnumerator];
    NSString	 *gsdocfile;

    while ((gsdocfile = [gsdocEnumerator nextObject]))
        {
        CREATE_AUTORELEASE_POOL (inner_pool);
        NSString	*texfile;
        NSString	*generated;

        texfile = [theFileManager texFileForGsdocFile: gsdocfile];

        if (options.ignoreDependencies || firstNewerThanSecond (gsdocfile, texfile))
            {
            GSXMLNode	*root;
            AGSTex      *tex;
            
            if (options.showDependencies == YES)
                {
                NSLog(@"%@ -> %@  ==> regenerate", gsdocfile, texfile);
                }

            tex = [AGSTex new];
            [tex setGlobalRefs: objects.globalRefs];
            [tex setProjectRefs: objects.projectRefs];

            root = [tex rootNodeAfterSetupFromFile: gsdocfile];

            generated = [tex outputDocument: root];

            if ([generated writeToFile: texfile atomically: YES] == NO)
                {
                NSLog(@"Sorry unable to write %@", texfile);
                }
            RELEASE (tex);
            }
        DESTROY (inner_pool);
        }
    DESTROY (arp);
}

void createXLPOutput (void)
{
    CREATE_AUTORELEASE_POOL (arp);
    
    NSEnumerator *gsdocEnumerator = [theFileManager gsdocFileEnumerator];
    NSString	 *gsdocfile;

    // 1 - write out the individual files corresponding to the gsdoc input
    
    while ((gsdocfile = [gsdocEnumerator nextObject]))
        {
        CREATE_AUTORELEASE_POOL (inner_pool);
        NSString	*xlpfile;
        NSString	*generated;

        xlpfile = [theFileManager xlpFileForGsdocFile: gsdocfile];

        if (options.ignoreDependencies || firstNewerThanSecond (gsdocfile, xlpfile))
            {
            GSXMLNode	*root;
            AGSXlp      *xlp;
            
            if (options.showDependencies == YES)
                {
                NSLog(@"%@ -> %@  ==> regenerate", gsdocfile, xlpfile);
                }

            xlp = [AGSXlp new];
            [xlp setGlobalRefs: objects.globalRefs];
            [xlp setProjectRefs: objects.projectRefs];

            root = [xlp rootNodeAfterSetupFromFile: gsdocfile];

            generated = [xlp outputDocument: root];
            
            if ([generated writeToFile: xlpfile atomically: YES] == NO)
                {
                NSLog(@"Sorry unable to write %@", xlpfile);
                }
            RELEASE (xlp);
            }
        DESTROY (inner_pool);
        }
    // 2 - write out the main.xlp so HelpViewer can read these files
    {
    AGSXlp *xlp = [AGSXlp new];
    NSString *main;
    NSString *fileName;
    
    [xlp setGlobalRefs: objects.globalRefs];
    [xlp setProjectRefs: objects.projectRefs];

    main = [xlp tableOfContents];

    fileName = [theFileManager fileInDocumentationDirectory: @"main.xlp"];
    
    if ([main writeToFile: fileName atomically: YES] == NO)
        {
        NSLog (@"Sorry unable to write %@", fileName);
        }
    RELEASE (xlp);
    }
    DESTROY (arp);
}


void updateCrossReferencesForHTML (void)
{
    CREATE_AUTORELEASE_POOL (arp);

    NSEnumerator *htmlFileEnumerator = [theFileManager htmlFileEnumerator];
    NSString *htmlFile;

    NSFileManager *mgr;

    mgr = [NSFileManager defaultManager];
    
    while ((htmlFile = [htmlFileEnumerator nextObject]))
        {
        CREATE_AUTORELEASE_POOL (inner_pool);

        NSString	*src;
        NSString	*dst;

        src = [htmlFile lastPathComponent]; //CHECKME, what is going on here?
        dst = htmlFile;

        /*
         * If we can't find the file in the source directory, assume
         * it is in the documentation directory already, and just needs
         * cross-refs rebuilding.
         */
        if ([mgr isReadableFileAtPath: src] == NO)
            {
            src = dst;
            }

        if ([mgr isReadableFileAtPath: src] == YES)
            {
            NSData		*d;
            NSMutableString	*s;
            NSRange		r;
            unsigned		l;
            unsigned		p;
            AGSHtml		*html;

            html = AUTORELEASE([AGSHtml new]);
            [html setGlobalRefs: objects.globalRefs];
            [html setProjectRefs: objects.projectRefs];
            [html setLocalRefs: nil];

            s = [NSMutableString stringWithContentsOfFile: src];
            l = [s length];
            p = 0;
            r = NSMakeRange(p, l);
            r = [s rangeOfString: @"<a rel=\"gsdoc\" href=\""
                   options: NSLiteralSearch
                   range: r];
            while (r.length > 0)
                {
                NSRange	replace;
                NSString	*repstr;
                NSString	*href;
                NSString	*type;
                NSString	*unit = nil;

                replace.location = r.location;
                p = NSMaxRange(r);

                r = [s rangeOfString: @"\">"
                       options: NSLiteralSearch
                       range: NSMakeRange(p, l - p)];
                if (r.length == 0)
                    {
                    NSLog(@"Unterminated gsdoc rel at %u", p);
                    break;
                    }
                else
                    {
                    replace = NSMakeRange(replace.location,
                                          NSMaxRange(r) - replace.location);
                    href = [s substringWithRange:
                                  NSMakeRange(p, r.location - p)];
                    p = NSMaxRange(replace);
                    }

                /*
                 * Skip past the '#' to the local reference.
                 */
                r = [href rangeOfString: @"#"
                          options: NSLiteralSearch];
                if (r.length == 0)
                    {
                    NSLog(@"Missing '#' in href at %u", replace.location);
                    break;
                    }
                href = [href substringFromIndex: NSMaxRange(r)];

                /*
                 * Split out the reference type information.
                 */
                r = [href rangeOfString: @"$"
                          options: NSLiteralSearch];
                if (r.length == 0)
                    {
                    NSLog(@"Missing '$' in href at %u", replace.location);
                    break;
                    }
                type = [href substringToIndex: r.location];
                href = [href substringFromIndex: NSMaxRange(r)];

                /*
                 * Parse unit name from method or instance variable link.
                 */
                if ([type isEqual: @"method"] == YES
                    || [type isEqual: @"ivariable"] == YES)
                    {
                    if ([type isEqual: @"method"] == YES)
                        {
                        r = [href rangeOfString: @"-"
                                  options: NSLiteralSearch];
                        if (r.length == 0)
                            {
                            r = [href rangeOfString: @"+"
                                      options: NSLiteralSearch];
                            }
                        if (r.length > 0)
                            {
                            unit = [href substringToIndex: r.location];
                            href = [href substringFromIndex: NSMaxRange(r)-1];
                            }
                        }
                    else
                        {
                        r = [href rangeOfString: @"*"
                                  options: NSLiteralSearch];
                        if (r.length > 0)
                            {
                            unit = [href substringToIndex: r.location];
                            href = [href substringFromIndex: NSMaxRange(r)];
                            }
                        }
                    if (unit == nil)
                        {
                        NSLog(@"Missing unit name terminator at %u",
                              replace.location);
                        break;
                        }
                    }
                if (unit == nil)
                    {
                    repstr = [html makeLink: href
                                   ofType: type
                                   isRef: YES];
                    }
                else
                    {
                    repstr = [html makeLink: href
                                   ofType: type
                                   inUnit: unit
                                   isRef: YES];
                    }
                if (options.isVerbose)
                    {
                    NSLog(@"Replace %@ with %@",
                          [s substringWithRange: replace],
                          repstr ? repstr : @"self");
                    }
                if (repstr != nil)
                    {
                    int	offset = [repstr length] - replace.length;

                    p += offset;
                    l += offset;
                    [s replaceCharactersInRange: replace withString: repstr];
                    }

                r = [s rangeOfString: @"<a rel=\"gsdoc\" href=\""
                       options: NSLiteralSearch
                       range: NSMakeRange(p, l - p)];
                }

            d = [s dataUsingEncoding: NSUTF8StringEncoding];
            [d writeToFile: dst atomically: YES];
            }
        else 
            {
            NSLog(@"No readable documentation at '%@' ... skipping", src);
            }
        DESTROY (inner_pool);
        }
    DESTROY (arp);
}
//==============================================================================================


int
main(int argc, char **argv, char **env)
{
    CREATE_AUTORELEASE_POOL (arp);

#ifdef GS_PASS_ARGUMENTS
    [NSProcessInfo initializeWithArguments: argv count: argc environment: env];
#endif

    setOptionsToDefaultValues ();
    initializeOptionsFromUserDefaults ();
  
    theFileManager = [[FileManager alloc] initForProject: options.project];

    // ---------- 2. Load old index information ---------------------------------------------------
    /*
     * Load any old project indexing information.
     */
  
    {
    NSDictionary *originalIndex;
  
    objects.projectRefs = [AGSIndex new];
    originalIndex = [theFileManager contentsOfReferenceFile];

    if (originalIndex)
        {
        [objects.projectRefs mergeRefs: originalIndex override: NO];
        }
    }
  
    //------ if clean option specified, only clean and exit  ------------------
  
    if (options.clean)
        {
        clean ();
        return 0;
        }

    //-------------------------------------------------------------------------
    if (![theFileManager sourceFileEnumerator]
        && ![theFileManager gsdocFileEnumerator]
        && ![theFileManager htmlFileEnumerator])
        {
        NSLog(@"No filename arguments found ... giving up");
        return 1;
        }

    //-------------------------------------------------------------------------

    setOutputOptionsToDefaultValues ();
    initializeOutputOptionsFromUserDefaults ();

    processSourceFiles ();
    processGSDocFiles ([theFileManager contentsOfReferenceFile], [theFileManager timeStampOfReferenceFile]);
  
    objects.globalRefs = [AGSIndex new];
  
    /*
     * If we are either generating html output, or relocating existing
     * html documents, we must build up the indexing information needed
     * for any cross-referencing etc.
     */
    if (options.generateHtml == YES || [theFileManager htmlFileEnumerator])
        {
        createCrossReferences ();
        }

    /*
     * Next pass ... generate html output from gsdoc files if required.
     */
    if (options.generateHtml)
        {
        createHtmlOutput ();
        }

    if (options.generateTeX)
        {
        createTeXOutput ();
        }

    if (options.generateXLP)
        {
        createXLPOutput ();
        }
    
    /*
     * Relocate existing html documents if required ... adjust all cross
     * referencing within those documents.
     */
    updateCrossReferencesForHTML ();

    [theFileManager release];
    DESTROY (arp);
  
    return 0;
}
